import java.util.*;
import java.io.*;

class BioMechanics extends BioComponents {

 ///////////////////////////////////////////////////////////////////
 //Construction
 ///////////////////////////////////////////////////////////////////
  BioMechanics() {
    super();
  }

 ///////////////////////////////////////////////////////////////////
 //Locate Chip Boundary and
 //Determine BinXNumber, BInYNumber and PadBinNumber
 ///////////////////////////////////////////////////////////////////
  void LocateBed(double marginRatio) { //don't consider steiner points.
    System.out.println("$ LOCATE BED");
   //keep a copy of the previous bed size
    int oldChipSizeX=CellSize[0][0], oldChipSizeY=CellSize[0][1];
    char padSide[]; padSide=new char[TerminalLimit+1];
    for (int t=1; t<=TerminalNumber[0]; t++) {
      padSide[t]=WhichSideOfATerminalOnACell(0,t,1000);
      if (padSide[t]=='x') { System.out.println("ERROR. Cannot determine which side of pad " + TerminalName[0][t] + "!"); Sleep(); }
    }
   //Find Core Bounding Box
    int left = 999999999, right = -999999999, top = -999999999, bottom = 999999999;
    for (int c=1; c<=CellNumber; c++) {
      int l=CellPosition[c][0]-CellKeepOutThickness[c][0];
      int b=CellPosition[c][1]-CellKeepOutThickness[c][1];
      int r=CellPosition[c][0]+CellSize[c][0]+CellKeepOutThickness[c][2];
      int t=CellPosition[c][1]+CellSize[c][1]+CellKeepOutThickness[c][3];
      if (l<left) left=l;
      if (b<bottom) bottom=b;
      if (r>right) right=r;
      if (t>top) top=t;
    }
    //for (int s=1; s<=SilkNumber; s++) {
    //  int x=SteinerPosition[s][0], y=SteinerPosition[s][1];
    //  if (x<left) left=x;
    //  if (y<bottom) bottom=y;
    //  if (x>right) right=x;
    //  if (y>top) top=y;
    //}
   //calculate core size and position with margin.
    int coreCenterX=(left+right)/2, coreCenterY=(bottom+top)/2;
    CellSize[0][0]=(int)((right-left)*(1+2*marginRatio)); CellSize[0][1]=(int)((top-bottom)*(1+2*marginRatio));
   //check and correct aspect ratio
    double aspChip=((double)CellSize[0][1])/((double)CellSize[0][0]);
    if (aspChip>AspectRatioMaxRate) CellSize[0][0]=(int)(CellSize[0][1]/AspectRatioMaxRate);
    else if (aspChip<AspectRatioMinRate) CellSize[0][1]=(int)(CellSize[0][0]*AspectRatioMinRate);
   //determine the chip size according to the core-limited or pad-limited design.    
    int corePerimeter=2*(CellSize[0][0]+CellSize[0][1]);
    int padRingPerimeter=0;
    for (int t=0; t<=TerminalNumber[0]; t++) padRingPerimeter+=TerminalWidth[0][t]+PadSpacing;
    double padRingCoreRatio=((double)padRingPerimeter)/((double)corePerimeter);
    if (padRingCoreRatio>1.0) { //I/O-limited
      CellSize[0][0]=(int)(CellSize[0][0]*padRingCoreRatio); CellSize[0][1]=(int)(CellSize[0][1]*padRingCoreRatio);
      System.out.println("$  This is a pad-limited design. Chip size=" + CellSize[0][0] + "/" + CellSize[0][1]);
    }
   //Determine Chip 2D Bin and 1D Bin, this may change CellSize[0][0] and CellSize[0][1].
    DetermineBin();
   //Chip Position
    CellPosition[0][0]=coreCenterX-CellSize[0][0]/2; CellPosition[0][1]=coreCenterY-CellSize[0][1]/2;
   //re-adjust pad offset
    if (CellSize[0][0]==0 | CellSize[0][1]==0)  {
      System.out.println("ERROR. Undefined bed, which has zero size.");
      Sleep();
    }
    double chipSizeExpandRatioX=((double)CellSize[0][0])/((double)oldChipSizeX);
    double chipSizeExpandRatioY=((double)CellSize[0][1])/((double)oldChipSizeY);
    for (int t=1; t<=TerminalNumber[0]; t++) {
      switch (padSide[t]) {
        case 'l': TerminalOffset[0][t][0]=0;
                  TerminalOffset[0][t][1]=(int)(TerminalOffset[0][t][1]*chipSizeExpandRatioY); 
                  break;
        case 'r': TerminalOffset[0][t][0]=CellSize[0][0]; 
                  TerminalOffset[0][t][1]=(int)(TerminalOffset[0][t][1]*chipSizeExpandRatioY); 
                  break;
        case 'b': TerminalOffset[0][t][0]=(int)(TerminalOffset[0][t][0]*chipSizeExpandRatioX);
                  TerminalOffset[0][t][1]=0; 
                  break;
        case 't': TerminalOffset[0][t][0]=(int)(TerminalOffset[0][t][0]*chipSizeExpandRatioX);
                  TerminalOffset[0][t][1]=CellSize[0][1];
                  break;
        default : System.out.println("ERROR. Cannot determine which side of pad " + TerminalName[0][t] + "!"); Sleep();
      }
    }
   //push steiners back to the bed
    int bedLeft=CellPosition[0][0]+1, bedRight=CellPosition[0][0]+CellSize[0][0]-1;
    int bedBottom=CellPosition[0][1]+1, bedTop=CellPosition[0][1]+CellSize[0][1]-1;
    for (int s=1; s<=SilkNumber; s++) {
      if (SteinerPosition[s][0]<bedLeft) SteinerPosition[s][0]=bedLeft;
      else if (SteinerPosition[s][0]>bedRight) SteinerPosition[s][0]=bedRight;
      if (SteinerPosition[s][1]<bedBottom) SteinerPosition[s][1]=bedBottom;
      else if (SteinerPosition[s][1]>bedTop) SteinerPosition[s][1]=bedTop;
    }
   //report
    System.out.println("$  Core perimeter=" + corePerimeter + ", pad ring perimeter=" + padRingPerimeter
                       + " Chip size=" + CellSize[0][0] + "/" + CellSize[0][1] + ", position=" + CellPosition[0][0] + "/" + CellPosition[0][1]);
    System.out.println("$  Chip Bin X/Y number= " + BinXNumber + "/" + BinYNumber + " Bin size=" + BinSize);  
  }

  void DetermineBin() {
    double w=((double)CellSize[0][0])/BinNumberLimit, h=((double)CellSize[0][1])/BinNumberLimit;
    BinSize=(w>h)?((int)(w+1)):((int)(h+1));   //get the BinSize
    BinXNumber=CellSize[0][0]/BinSize+1; BinYNumber=CellSize[0][1]/BinSize+1;
    int newCell0SizeX=BinXNumber*BinSize, newCell0SizeY=BinYNumber*BinSize;
    CellSize[0][0]=newCell0SizeX; CellSize[0][1]=newCell0SizeY;
  }

 ///////////////////////////////////////////////////////////////////
 //2-Dimension Density & Filling
 ///////////////////////////////////////////////////////////////////
  void CalculateBinDensity() {
    double areaCell=0.0, areaSteiner=0.0;
   //clean bindensity, calculate the balance value.
    //clean up all bin density and set CellDensity intially be -1.0
    for (int bx=1; bx<=BinXNumber; bx++) {
      for (int by=1; by<=BinYNumber; by++) {
      	BinCellDensity[bx][by]=0.0; BinSilkDensity[bx][by]=0.0;
      }
    }
    //System.out.println("   Bins cleaned with BinX/YNumber " + BinXNumber + "/" + BinYNumber + " and BinSize " + BinSize);
   ///////////////////////////////////////////////////////////////////
   //calculate density by measuring the overlapping of a cell with bins
    System.out.println("$ CALCULATE CELL BIN DENSITY");
    int left, right, bottom, top;  //indicate which bin the cell's corner locates.
    int cellLeft, cellRight, cellBottom, cellTop;
    double marginLBX, marginLBY, marginRTX, marginRTY; //the margin thickness of the cell overlapping bins, should be double
    for (int c=1; c<=CellNumber; c++) {
      //calculate cell's corner with reference to boundary. KeepOutThickness is included.
      cellLeft=(CellPosition[c][0]-CellKeepOutThickness[c][0])-CellPosition[0][0];
      cellBottom=(CellPosition[c][1]-CellKeepOutThickness[c][1])-CellPosition[0][1];
      cellRight=(CellPosition[c][0]+CellSize[c][0]+CellKeepOutThickness[c][2])-CellPosition[0][0];
      cellTop=(CellPosition[c][1]+CellSize[c][1]+CellKeepOutThickness[c][3])-CellPosition[0][1];
      //accumulate cell area
      areaCell+=((double)(cellRight-cellLeft))*((double)(cellTop-cellBottom));
      //find which bins the cell corners locate. Assume no cell will stay outside the chip boundary.
      //notice that marginRTX/Y are negative, distance to next bin. 
      //subtle problem. Eg. cellLeft/BinSize=-0.5->=0, this will lead to left=1, but then marginLBX>1.0
      //so we need to test the sign of cellLeft before calculating left.
      if (cellLeft<0) { left=1; marginLBX=1.0; } else { left=cellLeft/BinSize+1; marginLBX=(left*BinSize-cellLeft)/((double)BinSize); }
      if (cellRight>CellSize[0][0]) { right=BinXNumber; marginRTX=0.0; } else { right=cellRight/BinSize+1; marginRTX=(cellRight-right*BinSize)/((double)BinSize); }
      if (cellBottom<0) { bottom=1; marginLBY=1.0; } else { bottom=cellBottom/BinSize+1; marginLBY=(bottom*BinSize-cellBottom)/((double)BinSize); }
      if (cellTop>CellSize[0][1]) { top=BinYNumber; marginRTY=0.0; } else { top=cellTop/BinSize+1; marginRTY=(cellTop-top*BinSize)/((double)BinSize); }
      //bin density calculation
      BinCellDensity[left][bottom]+=marginLBX*marginLBY; BinCellDensity[left][top]+=marginLBX*marginRTY;
      BinCellDensity[right][bottom]+=marginRTX*marginLBY; BinCellDensity[right][top]+=marginRTX*marginRTY;
      for (int bx=left+1; bx<=right; bx++) {    //horizontally
        BinCellDensity[bx][bottom]+=marginLBY;                                  //bottom row
        for (int by=bottom+1; by<=top; by++) BinCellDensity[bx][by]+=1.0;       //in-between rows
        BinCellDensity[bx][top]+=marginRTY;                                     //top row
      }
      for (int by=bottom+1; by<=top; by++) {    //vertically
        BinCellDensity[left][by]+=marginLBX; BinCellDensity[right][by]+=marginRTX;   //left and right column
      }
      //store in Cell's fields
      CellBinBound[c][0]=left; CellBinBound[c][1]=bottom; CellBinBound[c][2]=right; CellBinBound[c][3]=top;
      CellBinBoundMargin[c][0]=marginLBX; CellBinBoundMargin[c][1]=marginLBY;
      CellBinBoundMargin[c][2]=marginRTX; CellBinBoundMargin[c][3]=marginRTY;
      //System.out.println("   " + CellName[c] + " margin " + marginLBX + " " + marginLBY + " " + marginRTX + " " + marginRTY);
    }
   ///////////////////////////////////////////////////////////////////
   //calculate Silk Bin Density by measuring which bin the STAR is in
    System.out.println("$ CALCULATE SILK BIN DENSITY");
    for (int s=1; s<=SilkNumber; s++) {
      SteinerBin[s][0]=(SteinerPosition[s][0]-CellPosition[0][0])/BinSize+1; 
      SteinerBin[s][1]=(SteinerPosition[s][1]-CellPosition[0][1])/BinSize+1;
      //find the bounding box for a silk including all its terminals and the steiner.
      //also annotate TerminalBin
      int binBoundaryLBX=SteinerBin[s][0], binBoundaryLBY=SteinerBin[s][1], binBoundaryRTX=SteinerBin[s][0], binBoundaryRTY=SteinerBin[s][1];
      for (int e=1; e<=SilkTerminalNumber[s]; e++) {
      	int c=SilkTerminalCell[s][e], t=SilkTerminalTerminal[s][e];
      	int terminalX=TerminalPosition[c][t][0]-CellPosition[0][0], terminalY=TerminalPosition[c][t][1]-CellPosition[0][1];  //relative to chip left-bottom corner
      	int terminalBinX=terminalX/BinSize+1, terminalBinY=terminalY/BinSize+1;
      	if (terminalBinX>BinXNumber) terminalBinX=BinXNumber; //thr right-side pad will otherwise result in terminalBinX=BinXNumber+1
      	if (terminalBinY>BinYNumber) terminalBinY=BinYNumber; //thr top-side pad will otherwise result in terminalBinY=BinYNumber+1
      	TerminalBin[c][t][0]=terminalBinX; TerminalBin[c][t][1]=terminalBinY;
      	if (terminalBinX<binBoundaryLBX) binBoundaryLBX=terminalBinX;
      	else if (terminalBinX>binBoundaryRTX) binBoundaryRTX=terminalBinX;
      	if (terminalBinY<binBoundaryLBY) binBoundaryLBY=terminalBinY;
      	else if (terminalBinY>binBoundaryRTY) binBoundaryRTY=terminalBinY;
      }	
      //temporary bin structure, preassumably cleared to zero when defined. Just for the bins within the bin boundary gotten above.
      double BinSilkPossessingProbability[][]; int BinSilkPossessingNumber[][];
      int binBoundaryNumberX=binBoundaryRTX-binBoundaryLBX+1, binBoundaryNumberY=binBoundaryRTY-binBoundaryLBY+1;
      BinSilkPossessingProbability=new double[binBoundaryNumberX][binBoundaryNumberY];   //notice that index from 0 to binBoundaryNumber-1
      BinSilkPossessingNumber=new int[binBoundaryNumberX][binBoundaryNumberY];  
      //System.out.println("$   " + SilkName[s] + " steinerBin " + SteinerBin[s][0] + "/" + SteinerBin[s][1]
      //                   + " bbox " + binBoundaryLBX + "/" + binBoundaryLBY + "/" + binBoundaryRTX + "/" + binBoundaryRTY);
      //working on the temporary bin structure.
      int sBinX=SteinerBin[s][0]-binBoundaryLBX, sBinY=SteinerBin[s][1]-binBoundaryLBY;
      for (int e=1; e<=SilkTerminalNumber[s]; e++) {
        int tBinX=TerminalBin[SilkTerminalCell[s][e]][SilkTerminalTerminal[s][e]][0]-binBoundaryLBX;
        int tBinY=TerminalBin[SilkTerminalCell[s][e]][SilkTerminalTerminal[s][e]][1]-binBoundaryLBY;
      	//calculate the probablity of silk taking a bin.
      	int T2SNumberX=java.lang.Math.abs(tBinX-sBinX)+1, T2SNumberY=java.lang.Math.abs(tBinY-sBinY)+1;
      	int dT2SY=(tBinY>sBinY)?(-1):1, dT2SX=(tBinX>sBinX)?(-1):1;
        //System.out.println("     " + TerminalName[SilkTerminalCell[s][e]][SilkTerminalTerminal[s][e]]
        //   + " TerminalBinX/Y " + TerminalBin[SilkTerminalCell[s][e]][SilkTerminalTerminal[s][e]][0] + "/"
        //   + TerminalBin[SilkTerminalCell[s][e]][SilkTerminalTerminal[s][e]][1]
        //   + " tBinX/Y " + tBinX + "/" + tBinY + " dT2SX/Y " + dT2SX + "/" + dT2SY + " T2SNumberX/Y " + T2SNumberX + "/" + T2SNumberY);
      	for (int iy=1,by=tBinY; iy<T2SNumberY; iy++) {
      	  for (int ix=1,bx=tBinX; ix<T2SNumberX; ix++) {
            BinSilkPossessingProbability[bx][by]+=BinProbabilityTable[ix][iy];
            bx+=dT2SX;
          }
          by+=dT2SY;
        }
      	for (int ix=1,bx=tBinX; ix<T2SNumberX; ix++) {
          BinSilkPossessingProbability[bx][sBinY]+=BinHorizontalEndEdgeProbabilityTable[ix][T2SNumberY];
          bx+=dT2SX;
        }
      	for (int iy=1,by=tBinY; iy<T2SNumberY; iy++) {
          BinSilkPossessingProbability[sBinX][by]+=BinVerticalEndEdgeProbabilityTable[T2SNumberX][iy];
          by+=dT2SY;
        }
        BinSilkPossessingProbability[sBinX][sBinY]+=1.0;
        //accumulate number of silks that possess the same bin.
        for (int iy=1, by=tBinY; iy<=T2SNumberY; iy++) {
      	  for (int ix=1, bx=tBinX; ix<=T2SNumberX; ix++) {
       	    BinSilkPossessingNumber[bx][by]++; 
       	    bx+=dT2SX;
          }
          by+=dT2SY;
        }
      }
      //average binPossessingProbability by binPossessingNumber
      for (int by=0; by<binBoundaryNumberY; by++) {
        //String aLine=new String();
        for (int bx=0; bx<binBoundaryNumberX; bx++) {
          if (BinSilkPossessingNumber[bx][by]!=0) BinSilkPossessingProbability[bx][by]/=BinSilkPossessingNumber[bx][by];
          //aLine+=" " + BinSilkPossessingProbability[bx][by];
        }
        //System.out.println("     " + by + ": " + aLine);
      }
      //OK, annotate the BinSilkDensity for this silk.
      //annotate
      for (int by=0; by<binBoundaryNumberY; by++) {
        for (int bx=0; bx<binBoundaryNumberX; bx++) {
          BinSilkDensity[binBoundaryLBX+bx][binBoundaryLBY+by]+=BinSilkPossessingProbability[bx][by];
        }
      }
      //next silk
    }
    //normalize the silkBinDensity with respect to WirePitch, WireLayerNumber and BinSize.
    double aWireContributeToDensity=((double)WirePitch)/((double)BinSize)/((double)WireLayerNumber);
    for (int bx=1; bx<=BinXNumber; bx++) {
      for (int by=1; by<=BinYNumber; by++) {
      	BinSilkDensity[bx][by]*=aWireContributeToDensity;
      }
    }
    
  }

  double BinAreaAverageDensity(int lbX, int lbY, int rtX, int rtY) {   //calculate the average bin density of a rect area.
    double aD=0;
    for (int bx=lbX; bx<=rtX; bx++) {
      for (int by=lbY; by<=rtY; by++) aD+=BinCellDensity[bx][by];
    }
    return aD/((rtX-lbX+1)*(rtY-lbY+1));
  }

  void CalculateBinFillingForce_CLASSICAL() {   //including all other cells' influence no matter how far they are.
    double FillingForceFactor = 1.0;
    double fillingforcefactor = FillingForceFactor*BinSize;
    //long totalCellArea=0;
    //for (int c=1; c<=CellNumber; c++) {
    //  totalCellArea+=((long)CellSize[c][0])*((long)CellSize[c][1]);
    //}
    //long chipArea=((long)CellSize[0][0])*((long)CellSize[0][1]);
    //double balancedCellDensity=((double)totalCellArea)/((double)chipArea);
    System.out.println("$ CALCULATE CELL BIN FILLING FORCE AND SILK BIN FILLING FORCE");
    //System.out.println("$   totalCellArea=" + totalCellArea + " chipArea=" + chipArea + " balancedCellDensity=" + balancedCellDensity);
    for (int bx=1; bx<=BinXNumber; bx++) {
      for (int by=1; by<=BinYNumber; by++) {
        double bcffx=0.0, bcffy=0.0, bsffx=0.0, bsffy=0.0;
        for (int bbx=1; bbx<=BinXNumber; bbx++) {
          for (int bby=1; bby<=BinYNumber; bby++) {
            if (bbx==bx & bby==by) continue;  //don't let it consider itself, otherwise will divide by zero
            double distanceSquare=EuclideanDistanceSquare(bbx,bby,bx,by);
            double xComponent=(bx-bbx)/distanceSquare, yComponent=(by-bby)/distanceSquare;
            double offsettedBinDensity=BinCellDensity[bbx][bby]-1.0;
            bcffx+=offsettedBinDensity*xComponent; bcffy+=offsettedBinDensity*yComponent;
            //in calculating silkBin fillingForce, we ignore all non-congested bins.
            if (BinSilkDensity[bbx][bby]>1.0) { 
              double offsettedSilkBinDensity=BinSilkDensity[bbx][bby]-1.0;
              bsffx+=offsettedSilkBinDensity*xComponent; bsffy+=offsettedSilkBinDensity*yComponent; 
            }
          }
        }
        BinCellFillingForce[bx][by][0]=bcffx*fillingforcefactor; BinCellFillingForce[bx][by][1]=bcffy*fillingforcefactor;
        BinSilkFillingForce[bx][by][0]=bsffx*fillingforcefactor; BinSilkFillingForce[bx][by][1]=bsffy*fillingforcefactor;
      }
    }      
  }


 ///////////////////////////////////////////////////////////////////
 //Cell, Steiner Filling Force Calculation
 ///////////////////////////////////////////////////////////////////
  void CalculateCellFillingForce() {
    System.out.println("$ CALCULATE CELL FILLING FORCE");
    for (int c=1; c<=CellNumber; c++) {
      double cffx=0.0, cffy=0.0;
      int left=CellBinBound[c][0], bottom=CellBinBound[c][1], right=CellBinBound[c][2], top=CellBinBound[c][3];
      double marginLBX=CellBinBoundMargin[c][0], marginLBY=CellBinBoundMargin[c][1],
             marginRTX=CellBinBoundMargin[c][2], marginRTY=CellBinBoundMargin[c][3];
     //cell corners
      cffx+=BinCellFillingForce[left][bottom][0]*marginLBX*marginLBY; cffy+=BinCellFillingForce[left][bottom][1]*marginLBX*marginLBY;  //left-bottom corner
      cffx+=BinCellFillingForce[left][top][0]*marginLBX*marginRTY; cffy+=BinCellFillingForce[left][top][1]*marginLBX*marginRTY;  //left-top corner
      cffx+=BinCellFillingForce[right][bottom][0]*marginRTX*marginLBY; cffy+=BinCellFillingForce[right][bottom][1]*marginRTX*marginLBY;  //right-bottom corner
      cffx+=BinCellFillingForce[right][top][0]*marginRTX*marginRTY; cffy+=BinCellFillingForce[right][top][1]*marginRTX*marginRTY;  //right-top corner
     //cell edges and cell central area
      for (int bx=left+1; bx<=right; bx++) {    //horizontally
        cffx+=BinCellFillingForce[bx][bottom][0]*marginLBY; cffy+=BinCellFillingForce[bx][bottom][1]*marginLBY;              //bottom row
        for (int by=bottom+1; by<=top; by++) { cffx+=BinCellFillingForce[bx][by][0]; cffy+=BinCellFillingForce[bx][by][1]; } //in-between rows
        cffx+=BinCellFillingForce[bx][top][0]*marginRTY; cffy+=BinCellFillingForce[bx][top][1]*marginRTY;                    //top row
      }
      for (int by=bottom+1; by<=top; by++) {    //vertically
        cffx+=BinCellFillingForce[left][by][0]*marginLBX; cffy+=BinCellFillingForce[left][by][1]*marginLBY;     //left column
        cffx+=BinCellFillingForce[right][by][0]*marginRTX; cffy+=BinCellFillingForce[right][by][1]*marginRTY;    //right column
      }
     //annotate CellFillingForce, they should be averaged thru all bins the cell covered.
      double cellAreaBinRatio=((right+1+marginRTX)-(left+1-marginLBX))*((top+1+marginRTY)-(bottom+1-marginLBY));
      //System.out.println(" BinSize=" + BinSize + ", Cell " + CellName[c] + " cellAreaBinRatio=" + cellAreaBinRatio);
      CellFillingForce[c][0]=(int)(cffx/cellAreaBinRatio); CellFillingForce[c][1]=(int)(cffy/cellAreaBinRatio);
    }
  }

  void CalculateTerminalSteinerFillingForce() {
    System.out.println("$ CALCULATE TERMINAL AND STEINER FILLING FORCE");
    for (int s=1; s<=SilkNumber; s++) {
      int sbx=SteinerBin[s][0], sby=SteinerBin[s][1];
      double sfx=0.0, sfy=0.0;
      int sfNum=0;
      for (int e=1; e<=SilkTerminalNumber[s]; e++) {
      	int c=SilkTerminalCell[s][e], t=SilkTerminalTerminal[s][e];
      	int tbx=TerminalBin[c][t][0], tby=TerminalBin[c][t][1];
      	double tfx=0.0, tfy=0.0;
      	int tfNum=0;
      	int t2sXDir=1, t2sYDir=1;
      	int t2sXNumber=1+java.lang.Math.abs(sbx-tbx), t2sYNumber=1+java.lang.Math.abs(sby-tby);
      	if (t2sXNumber==1) t2sXDir=0; else if (tbx>sbx) t2sXDir=-1;
      	if (t2sYNumber==1) t2sYDir=0; else if (tby>sby) t2sYDir=-1;
      	for (int i=1,bx=tbx; i<=t2sXNumber; i++,bx+=t2sXDir) {
      	  int j4T=(t2sXNumber==1)?(t2sYNumber/2):((int)(t2sYNumber*(1-(i-1)/((double)t2sXNumber))+0.5));
      	  for (int j=1,by=tby; j<=t2sYNumber; j++,by+=t2sYDir) {
      	    if (j<=j4T) {
              tfx+=BinSilkFillingForce[bx][by][0]; tfy+=BinSilkFillingForce[bx][by][1]; tfNum++;
      	    }
      	    else {
              sfx+=BinSilkFillingForce[bx][by][0]; sfy+=BinSilkFillingForce[bx][by][1]; sfNum++;
            }
          }
        }
        if (tfNum!=0) { TerminalFillingForce[c][t][0]=(int)(tfx/tfNum); TerminalFillingForce[c][t][1]=(int)(tfy/tfNum); }
        else { TerminalFillingForce[c][t][0]=0; TerminalFillingForce[c][t][1]=0; }
      }
      if (sfNum!=0) { SteinerFillingForce[s][0]=(int)(sfx/sfNum); SteinerFillingForce[s][1]=(int)(sfy/sfNum); }
      else { SteinerFillingForce[s][0]=0; SteinerFillingForce[s][1]=0; }
    }
  }

///////////////////////////////////////////////////////////////////////////////////////
// ALL ABOUT CHAIN
// Created on 001003
// Modified on 001009
///////////////////////////////////////////////////////////////////////////////////////
  boolean Finished(int a) {
    if (a<=150) return true; else return false;
  }

  void CalculateArgument(int a) {  
    //Terminal: WeightChainAttractForce+WeightSilkAttractForce+WeightRoutingFillingForce=1
    //Cell:     WeightCellAttractForce+WeightCellFillingForce=1
    //Steiner:  WeightSteinerAttractForce+WeightSteinerFillingForce=1
    //
    //TerminalForce=WeightChainAttractForce*TerminalChainAttractForce
    //             +WeightSilkAttractForce*TerminalSIlkSilkAttractForce
    //             +WeightRoutingFillingForce*TerminalFillingForce
    //CellForce=WeightCellAttractForce*CellAttractForce+WeightCellFillingForce*CellFillingForce
    //SteinerForce=WeightSteinerAttractForce*SteinerAttractForce+WeightSteinerFillingForce*SteinerFillingForce
    //
    //CellMoveLimit, SteinerMoveLimit
    //CellMoveLimit=(java.lang.Math.min(CellSize[0][0],CellSize[0][1]))/10;
    //SteinerMoveLimit=(java.lang.Math.min(CellSize[0][0],CellSize[0][1]))/10;
    //
    //CellNewRatio
    CellNewRatio=(a>100)?(1.0):(0.01*a);
    WeightRoutingFillingForce=0.1;
    WeightChainAttractForce=0.0; 
    WeightSilkAttractForce=1.0-WeightChainAttractForce-WeightRoutingFillingForce;
    if (a<=100) WeightCellFillingForce=0.01*a;
    WeightCellAttractForce=1.0-WeightCellFillingForce;
    WeightSteinerAttractForce=0.9; WeightSteinerFillingForce=1.0-WeightSteinerAttractForce;
    CellMoveLimit=(int)((java.lang.Math.min(CellSize[0][0],CellSize[0][1]))/(1.0+a*0.4));
    if (a>=100) CellMoveLimit=2*BinSize;
    SteinerMoveLimit=(int)((java.lang.Math.min(CellSize[0][0],CellSize[0][1]))/(1.0+a*0.01));
    System.out.println("$ Weight of Silk/Chain/Routing=" + WeightSilkAttractForce + "/" + WeightChainAttractForce + "/" + WeightRoutingFillingForce);
    System.out.println("$ Weight of Cell Attract/Filling=" + WeightCellAttractForce + "/" + WeightCellFillingForce);
    System.out.println("$ Weight of Steiner Attract/Filling=" + WeightSteinerAttractForce + "/" + WeightSteinerFillingForce);
    System.out.println("$ CellNewRatio=" + CellNewRatio + " MoveLimit of Cell/Steiner=" + CellMoveLimit + "/" + SteinerMoveLimit);
  }
  
  void CalculateSilkLength() {
    System.out.println("$ CALCULATE SILK LENGTH");
    for (int c=0; c<=CellNumber; c++) {
      for (int t=1; t<=TerminalNumber[c]; t++) {
      	int tX=CellPosition[c][0]+TerminalOffset[c][t][0], tY=CellPosition[c][1]+TerminalOffset[c][t][1];
      	TerminalPosition[c][t][0]=tX; TerminalPosition[c][t][1]=tY;
      	int sX=SteinerPosition[TerminalSilk[c][t]][0], sY=SteinerPosition[TerminalSilk[c][t]][1];
      	TerminalSilkLength[c][t]=EuclideanDistance(tX,tY,sX,sY);
      	if (TerminalSilkLength[c][t]!=0) {
          TerminalSilkDirection[c][t][0]=((double)(sX-tX))/((double)TerminalSilkLength[c][t]);   //cos
          TerminalSilkDirection[c][t][1]=((double)(sY-tY))/((double)TerminalSilkLength[c][t]);   //sin
        }
        else { TerminalSilkDirection[c][t][0]=0.0; TerminalSilkDirection[c][t][1]=0.0; }
      }
    }
  }

  void CalculateSilkAttractForceValue() {
    System.out.println("$ CALCULATE SILK ATTRACT FORCE VALUE");
    //no need to clean first, because one silk one force
    for (int c=0;c<=CellNumber;c++) for (int t=1;t<=TerminalNumber[c];t++) TerminalSilkAttractForceValue[c][t]=TerminalSilkLength[c][t];
  }

  void CalculateFillingForce() {
    System.out.println("$ CALCULATE CELL/TERMINAL-STEINER FILLING FORCE");
    CalculateBinDensity();
    CalculateBinFillingForce_CLASSICAL();
    CalculateCellFillingForce();
    CalculateTerminalSteinerFillingForce();
  }

  void CalculateCompleteForce() {  //old name=CalculateCellSteinerCompleteForce, now including CalculateSilkLenght/SilkAttractForceVlaue/CellSteinerFillingForce.
    System.out.println("$ CALCULATE COMPLETE FORCE");
    //Terminal: WeightChainAttractForce+WeightSilkAttractForce+WeightRoutingFillingForce=1
    //Cell:     WeightCellAttractForce+WeightCellFillingForce=1
    //Steiner:  WeightSteinerAttractForce+WeightSteinerFillingForce=1
    //
    //TerminalForce=WeightChainAttractForce*TerminalChainAttractForce
    //             +WeightSilkAttractForce*TerminalSIlkSilkAttractForce
    //             +WeightRoutingFillingForce*TerminalFillingForce
    //CellForce=WeightCellAttractForce*CellAttractForce+WeightCellFillingForce*CellFillingForce
    //SteinerForce=WeightSteinerAttractForce*SteinerAttractForce+WeightSteinerFillingForce*SteinerFillingForce
    //
    CalculateSilkLength();
    CalculateSilkAttractForceValue();
    //CalculateChainAttractForceValue();
    CalculateFillingForce();
    //clean all attract force
    for (int s=1; s<=SilkNumber; s++) { SteinerAttractForce[s][0]=0; SteinerAttractForce[s][1]=0; }
    for (int c=0; c<=CellNumber; c++) { CellAttractForce[c][0]=0; CellAttractForce[c][1]=0; }
    //computation
    for (int c=0; c<=CellNumber; c++) {
      for (int t=1; t<=TerminalNumber[c]; t++) {
      	double fATValue=WeightChainAttractForce*TerminalChainAttractForceValue[c][t]
      	                +WeightSilkAttractForce*TerminalSilkAttractForceValue[c][t];  //weighted terminal attract force value
      	int fx=(int)(fATValue*TerminalSilkDirection[c][t][0]), fy=(int)(fATValue*TerminalSilkDirection[c][t][1]);
      	SteinerAttractForce[TerminalSilk[c][t]][0]-=fx;
      	SteinerAttractForce[TerminalSilk[c][t]][1]-=fy;
      	TerminalForce[c][t][0]=fx+(int)(WeightRoutingFillingForce*TerminalFillingForce[c][t][0]); 
      	TerminalForce[c][t][1]=fy+(int)(WeightRoutingFillingForce*TerminalFillingForce[c][t][1]);
      	CellAttractForce[c][0]+=TerminalForce[c][t][0];
      	CellAttractForce[c][1]+=TerminalForce[c][t][1];
      }
      CellAttractForce[c][0]/=TerminalNumber[c];
      CellAttractForce[c][1]/=TerminalNumber[c];
      CellForce[c][0]=(int)(WeightCellAttractForce*CellAttractForce[c][0]+WeightCellFillingForce*CellFillingForce[c][0]); 
      CellForce[c][1]=(int)(WeightCellAttractForce*CellAttractForce[c][1]+WeightCellFillingForce*CellFillingForce[c][1]);
    }
    for (int s=1; s<=SilkNumber; s++) {
      SteinerAttractForce[s][0]/=SilkTerminalNumber[s];
      SteinerAttractForce[s][1]/=SilkTerminalNumber[s];
      SteinerForce[s][0]=(int)(WeightSteinerAttractForce*SteinerAttractForce[s][0]+WeightSteinerFillingForce*SteinerFillingForce[s][0]);
      SteinerForce[s][1]=(int)(WeightSteinerAttractForce*SteinerAttractForce[s][1]+WeightSteinerFillingForce*SteinerFillingForce[s][1]);
    }
  }

  void CalculateCellPosition() {
    System.out.println("$ CALCULATE CELL POSITION");
    double limit=(double)CellMoveLimit;
    for (int c=1; c<=CellNumber; c++) {
      double fx=(double)CellForce[c][0], fy=(double)CellForce[c][1];
      double sf=java.lang.Math.abs(fx);
      if (sf>limit) {
        double fRatio=(sf/limit);  //fRatio always > +1
        fx=fx/fRatio; fy=fy/fRatio;
      }
      sf=java.lang.Math.abs(fy);
      if (sf>limit) {
        double fRatio=(sf/limit);  //fRatio always > +1
        fx=fx/fRatio; fy=fy/fRatio;
      }
      CellPosition[c][0]+=(int)(fx); CellPosition[c][1]+=(int)(fy);
    }
  }

  void CalculateSteinerPosition() { 
    System.out.println("$ CALCULATE STEINER POSITION");
    double limit=(double)SteinerMoveLimit;
    for (int s=1; s<=SilkNumber; s++) {
      double fx=(double)SteinerForce[s][0], fy=(double)SteinerForce[s][1];
      fx/=SilkTerminalNumber[s]; fy/=SilkTerminalNumber[s];
      double sf=java.lang.Math.abs(fx);
      if (sf>limit) {
        double fRatio=(sf/limit);  //fRatio always > +1
        fx=fx/fRatio; fy=fy/fRatio;
      }
      sf=java.lang.Math.abs(fy);
      if (sf>limit) {
        double fRatio=(sf/limit);  //fRatio always > +1
        fx=fx/fRatio; fy=fy/fRatio;
      }
      SteinerPosition[s][0]+=(int)fx; SteinerPosition[s][1]+=(int)fy;
    }
  }



///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT
//SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT SOFT
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  void CalculateSoftCellShape() {
    System.out.println("$ CALCULATE SOFT CELL SHAPE");
    for (int c=1; c<=CellNumber; c++) if (CellType[c]=='c' | CellType[c]=='d') MassageASoftCell(c);
  }    
    
  void MassageASoftCell(int c) {
    int massageLimit=(int)(java.lang.Math.min(CellSize[c][0],CellSize[c][1])*0.1);
    int cpX=CellPosition[c][0], cpY=CellPosition[c][1];
    long cellArea=CellSize[c][0]*CellSize[c][1];
    double asp=((double)CellSize[c][1])/((double)CellSize[c][0]);
    double aspSQ=asp*asp;
    double QD=0, HH=0, HDC=0;
    for (int t=1; t<=TerminalNumber[c]; t++) {
      switch (WhichSideOfATerminalOnACell(c,t,100)) {
      	case 'l': HH-=(double)(cpX+TerminalOffset[c][t][0]-SteinerPosition[TerminalSilk[c][t]][0]);
      	          QD+=1;
      	          break;
      	case 'r': HH+=(double)(cpX+TerminalOffset[c][t][0]-SteinerPosition[TerminalSilk[c][t]][0]);
      	          QD+=1;
      	          break;
      	case 'b': HH+=asp*(cpY+TerminalOffset[c][t][1]-SteinerPosition[TerminalSilk[c][t]][1]);
      	          QD+=aspSQ;
      	          break;
      	case 't': HH-=asp*(cpY+TerminalOffset[c][t][1]-SteinerPosition[TerminalSilk[c][t]][1]);
      	          QD+=aspSQ;
      	          break;
      	default : break;
      }
    }
    int lbx=CellBinBound[c][0], lby=CellBinBound[c][1], rtx=CellBinBound[c][2], rty=CellBinBound[c][3];
    //left and right boundary bins.
    for (int by=lby; by<=rty; by++) HDC+=BinCellFillingForce[lbx][by][0]-BinCellFillingForce[rtx][by][0];
    //bottom and top boundary bins.
    for (int bx=lbx; bx<=rtx; bx++) HDC+=asp*(-BinCellFillingForce[bx][lby][1]+BinCellFillingForce[bx][rty][1]);
    //get optimum delta wx. Don't forget the minus ahead!
    double dwxOPT=-(HH+(WeightCellFillingForce/WeightCellAttractForce)*HDC)/(2*QD);
    //System.out.println("^  Cell " + CellName[c] + " size=" + CellSize[c][0] + "/" + CellSize[c][1] + " asp=" + asp);
    //System.out.println("^   HH=" + HH + " HDC=" + HDC + " QD=" + QD + " dwxOPT=" + dwxOPT);
    //check the validity.
    if (dwxOPT>massageLimit) dwxOPT=massageLimit;
    else if (dwxOPT<-massageLimit) dwxOPT=-massageLimit;
    double wxNew=CellSize[c][0]+2*dwxOPT;
    double wxMinLimit=java.lang.Math.sqrt(cellArea/CellAspectRatioMax[c]);
    double wxMaxLimit=java.lang.Math.sqrt(cellArea/CellAspectRatioMin[c]);
    if (wxNew<wxMinLimit) wxNew=wxMinLimit;
    else if (wxNew>wxMaxLimit) wxNew=wxMaxLimit;
    double wyNew=cellArea/wxNew;
    //scaling position, must do this before updating cellsize.
    CellPosition[c][0]-=(int)((wxNew-CellSize[c][0])/2);
    CellPosition[c][1]-=(int)((wyNew-CellSize[c][1])/2);
    //get new cell size
    CellSize[c][0]=(int)wxNew; CellSize[c][1]=(int)wyNew;
    //checking terminals, because after reshaping, some terminals may get out of cell boundary.
    //but this is not complete, 'cos some corner terminals may not be updated correctly.
    for (int t=1; t<=TerminalNumber[c]; t++) {
      //check if the terminal is outside the cell, if so correct.
      if (TerminalOffset[c][t][0]<0) TerminalOffset[c][t][0]=0;
      else if (TerminalOffset[c][t][0]>CellSize[c][0]) TerminalOffset[c][t][0]=CellSize[c][0];
      if (TerminalOffset[c][t][1]<0) TerminalOffset[c][t][1]=0;
      else if (TerminalOffset[c][t][1]>CellSize[c][1]) TerminalOffset[c][t][1]=CellSize[c][1];
    }
  }

 //////////////////////////////////////////////////////////////////////////////////////////////////

  void CalculateSoftTerminalPosition() {
    System.out.println("$ CALCULATE SOFT TERMINAL POSITION");
    for (int c=1; c<=CellNumber; c++) {
      switch(CellType[c]) {
      	case 'c': CalculateSoftBodyTerminalPosition(c); break;
      	case 'd': CalculateSoftBoundaryTerminalPosition(c); break;
      	default : break; 
      }	
    }
  }
  
  void CalculateSoftBodyTerminalPosition(int c) {
    if (CellType[c]!='c') System.out.println("!WARNING. Cell type of " + CellName[c] + ", " + CellType[c] + " is not a valid soft cell with soft-body terminals.");
    int csX=CellSize[c][0], csY=CellSize[c][1];
    //System.out.println("^  Cell " + CellName[c] + " size=" + csX + "/" + csY);
    for (int t=1; t<=TerminalNumber[c]; t++) {
      int tox=TerminalOffset[c][t][0], toy=TerminalOffset[c][t][1];
      int fx=TerminalForce[c][t][0], fy=TerminalForce[c][t][1];
      int toxNew=tox+fx, toyNew=toy+fy;
      //System.out.println("^   Terminal " + CellName[c] + "/" + TerminalName[c][t]
      //                   + " to=" + tox + "/" + toy + " toDes=" + toxNew + "/" + toyNew);
      if (toxNew<0 | toxNew>csX | toyNew<0 | toyNew>csY) {  
        for (int stepX=fx/2,stepY=fy/2; (java.lang.Math.abs(stepX)>0)|(java.lang.Math.abs(stepY)>0); stepX/=2,stepY/=2) { //binary search, ends with the terminal closer to cell boundary.
          toxNew=tox+fx; toyNew=toy+fy;  
          if (toxNew<0 | toxNew>csX | toyNew<0 | toyNew>csY) { fx-=stepX; fy-=stepY; } else { fx+=stepX; fy+=stepY; }
        }
      }
      TerminalOffset[c][t][0]=toxNew; TerminalOffset[c][t][1]=toyNew;
      //System.out.println("^     toNew=" + toxNew + "/" + toyNew);
      switch (WhichSideOfATerminalOnACell(c,t,1000)) {  // correct those terminals slightly outside the boundary.
      	case 'l': TerminalOffset[c][t][0]=0; break;
      	case 'r': TerminalOffset[c][t][0]=CellSize[c][0]; break;
      	case 'b': TerminalOffset[c][t][1]=0; break;
      	case 't': TerminalOffset[c][t][1]=CellSize[c][1]; break; 
      	default : break;
      }
    }  	
  }

  void CheckSoftBoundaryTerminal() {
    System.out.println("$  CHECK SOFT BOUNDARY TERMINAL");
    for (int c=0; c<=CellNumber; c++) CheckSoftBoundaryTerminal(c);
  }
 
  void CheckSoftBoundaryTerminal(int c) {
    Random random = new Random(1);
    if (CellType[c]=='b' | CellType[c]=='d') {
      int csX=CellSize[c][0], csY=CellSize[c][1];
      for (int t=1; t<=TerminalNumber[c]; t++) {
      	//if outside bound
      	if (TerminalOffset[c][t][0]<0) TerminalOffset[c][t][0]=0;
      	else if (TerminalOffset[c][t][0]>csX) TerminalOffset[c][t][0]=csX;
      	if (TerminalOffset[c][t][1]<0) TerminalOffset[c][t][1]=0;
      	else if (TerminalOffset[c][t][1]>csY) TerminalOffset[c][t][1]=csY;
      	//if not on bound
        switch (WhichSideOfATerminalOnACell(c,t,1000)) {
          case 'l': TerminalOffset[c][t][0]=0; break;
          case 'r': TerminalOffset[c][t][0]=CellSize[c][0]; break;
          case 'b': TerminalOffset[c][t][1]=0; break;
          case 't': TerminalOffset[c][t][1]=CellSize[c][1]; break;
          default : switch((int)(4*random.nextDouble())) {
          	      case 0: TerminalOffset[c][t][0]=0; break;
           	      case 1: TerminalOffset[c][t][0]=CellSize[c][0]; break;
           	      case 2: TerminalOffset[c][t][1]=0; break;
            	      case 3: TerminalOffset[c][t][1]=CellSize[c][1]; break;
            	      default : System.out.println("ERROR. Random number out of expected range."); Sleep(); break;
            	    }
            	    //System.out.println("$  Terminal " + CellName[c] + "/" + TerminalName[c][t] + " be moved to boundary.");
            	    break;
        }
      }   
    }
  }

  void DeterminePerimeterBin(int c) {
    //System.out.println("$  DETERMINE PERIMETER BIN");
    if (TerminalNumber[c]==0) { System.out.println("ERROR. Cell " + CellName[c] + " has zero terminals"); Sleep(); }
    int totalTerminalWidth=0;
    int perimeter=(CellSize[c][0]+CellSize[c][1])*2;
    for (int t=1; t<=TerminalNumber[c]; t++) totalTerminalWidth+=TerminalWidth[c][t];
    int averageTerminalWidth=(int)(totalTerminalWidth/TerminalNumber[c]*CellRatio[c]);
    PerimeterBinNumber[c]=java.lang.Math.max(perimeter/(averageTerminalWidth/2),6);           //lower bound. Fillingforce should have +-1/4 range
    PerimeterBinNumber[c]=java.lang.Math.min(PerimeterBinNumber[c],PerimeterBinNumberLimit);  //upper bound
    PerimeterBinSize[c]=perimeter/PerimeterBinNumber[c];
    PerimeterRatio[c]=((double)perimeter)/(PerimeterBinNumber[c]*PerimeterBinSize[c]);  //a number >=1.0
    System.out.println("$  Cell " + CellName[c] + " perimeter=" + perimeter + " ave terminal width(ratio-ed)=" + averageTerminalWidth 
                        + " bin number=" + PerimeterBinNumber[c] + " bin size=" + PerimeterBinSize[c] + " perimeter raio=" + PerimeterRatio[c]);
  }

  void CalculatePerimeterBinDensity(int c) {
    if (CellType[c]!='b' & CellType[c]!='d') System.out.println("!WARNING. Cell type of " + CellName[c] + ", " + CellType[c] + " is not a valid soft cell with soft-boundary terminals.");
    int realPerimeter=(CellSize[c][0]+CellSize[c][1])*2;
    int perimeterBinNumber=PerimeterBinNumber[c];
    int perimeterBinSize=PerimeterBinSize[c];
    int workingPerimeter=perimeterBinNumber*perimeterBinSize;
    double perimeterRatio=PerimeterRatio[c];
    if (perimeterBinNumber<=0) { System.out.println("!ERROR. Cell " + CellName[c] + " perimeter bin number=" + perimeterBinNumber); Sleep(); }
    if (perimeterBinSize<=0) { System.out.println("!ERROR. Cell " + CellName[c] + " perimeter bin size=" + perimeterBinSize); Sleep(); }
    if (perimeterRatio<=0.9 | perimeterRatio>=1.2) { System.out.println("!ERROR. Cell " + CellName[c] + " perimeter ratio=" + perimeterRatio); Sleep(); }
    //clean balance density, bin indexing from 0 to PerimeterBinNumber[c]-1
    for (int b=0; b<=PerimeterBinNumber[c]-1; b++) PerimeterBinDensity[c][b]=0;
    //compute the contribution to density by each terminal
    for (int t=1; t<=TerminalNumber[c]; t++) {
      //determine which bins the terminal cover.
      //NOTE. here left and right is not really left and right. Left=clockwise direction. Right=counter-clockwise direction
      //and cyclic.
      //indexing from 0 to perimeterBinNumber-1
      //the 0-st bin starts the left-bottom corner of the chip and goes counter-clockwise.
      //determine the left/right bin of the terminal
      int offset=(int)(HowManyOffsetOfATerminalOnACellPerimeter(c,t)*perimeterRatio);
      int terminalWidthHalf=(int)(TerminalWidth[c][t]*perimeterRatio*CellRatio[c]/2);
      int terminalLeft=(offset-terminalWidthHalf+workingPerimeter)%workingPerimeter;
      int terminalRight=(offset+terminalWidthHalf)%workingPerimeter;
      int leftBin=terminalLeft/perimeterBinSize, rightBin=terminalRight/perimeterBinSize;
      double leftMargin=((leftBin+1)*perimeterBinSize-terminalLeft)/((double)perimeterBinSize);
      double rightMargin=(terminalRight-(rightBin+1)*perimeterBinSize)/((double)perimeterBinSize);  //always non-positive.
      //System.out.println("^  Terminal " + CellName[c] + "/" + TerminalName[c][t] + " offset=" + offset 
      //                   + " left/rightBin=" + leftBin + "/" + rightBin + " margin left/right=" + leftMargin + "/" + rightMargin);
      PerimeterBinDensity[c][leftBin]+=leftMargin; PerimeterBinDensity[c][rightBin]+=rightMargin;
      if (leftBin<=rightBin) { //normal case
        for (int b=leftBin+1; b<=rightBin; b++) PerimeterBinDensity[c][b]+=1.0;
      }
      else {                   //the terminal just sits across origin
      	for (int b=leftBin+1; b<perimeterBinNumber; b++) PerimeterBinDensity[c][b]+=1.0;
        for (int b=0; b<=rightBin; b++) PerimeterBinDensity[c][b]+=1.0;
      }
      //record
      Terminal1DBinBound[c][t][0]=leftBin; Terminal1DBinBound[c][t][1]=rightBin;
      Terminal1DBinBoundMargin[c][t][0]=leftMargin; Terminal1DBinBoundMargin[c][t][1]=rightMargin;
    }
  }

  void CalculatePerimeterBinFillingForce(int c) {
    int perimeterBinNumber=PerimeterBinNumber[c];
    double perimeterBinSize=(double)PerimeterBinSize[c];
    for (int b=1; b<=perimeterBinNumber; b++) {
      double bff=0.0;
      for (int dd=1; dd<=perimeterBinNumber/4; dd++) {  //range covers half of perimeter
        double bDensityLeftHand=PerimeterBinDensity[c][(b-dd+perimeterBinNumber)%perimeterBinNumber];
        if (bDensityLeftHand>1.0) bff+=(bDensityLeftHand-1.0)/dd;
        double bDensityRightHand=PerimeterBinDensity[c][(b+dd)%perimeterBinNumber];
        if (bDensityRightHand>1.0) bff-=(bDensityRightHand-1.0)/dd;
      }
      PerimeterBinFillingForce[c][b]=bff*perimeterBinSize/2;
    }
  }

  void CalculateSoftBoundaryTerminalFillingForce(int c) {
    int perimeterBinNumber=PerimeterBinNumber[c];
    for (int t=1; t<=TerminalNumber[c]; t++) {
      double terminalWidthVsBinSize;
      int leftBin=Terminal1DBinBound[c][t][0], rightBin=Terminal1DBinBound[c][t][1];
      double leftMargin=Terminal1DBinBoundMargin[c][t][0], rightMargin=Terminal1DBinBoundMargin[c][t][1];
      //corners
      double tff=PerimeterBinFillingForce[c][leftBin]*leftMargin+PerimeterBinFillingForce[c][rightBin]*rightMargin;
      //between corners
      if (leftBin<=rightBin) { //normal case 
        terminalWidthVsBinSize=(rightBin-leftBin)+leftMargin+rightMargin;
        for (int b=leftBin+1; b<=rightBin; b++) tff+=PerimeterBinFillingForce[c][b];
      }
      else {                   //the terminal just sits across origin
        terminalWidthVsBinSize=(perimeterBinNumber-(leftBin+1))+leftMargin;
      	terminalWidthVsBinSize+=rightBin+rightMargin;
      	for (int b=leftBin+1; b<perimeterBinNumber; b++) tff+=PerimeterBinFillingForce[c][b];
      	for (int b=0; b<=rightBin; b++) tff+=PerimeterBinFillingForce[c][b];
      }
      //this one sentence is for debugging.
      if (terminalWidthVsBinSize<=0.0) { System.out.println("ERROR. Terminal " + CellName[c] + "/" + TerminalName[c][t] + " terminalWidthVsBinSize=" + terminalWidthVsBinSize); Sleep(); }
      Terminal1DFillingForce[c][t]=(int)(tff/(terminalWidthVsBinSize*PerimeterRatio[c]));
    }
  }

  void CalculateSoftBoundaryTerminal1DForce(int c) {  //the terminal force caused by attract force is its tangential part wrt the terminal-offset-to-cell-center.
    int cellCenterX=CellSize[c][0]/2, cellCenterY=CellSize[c][1]/2;
    double halfPerimeter=(double)(CellSize[c][0]+CellSize[c][1]);
    if (halfPerimeter==0) { System.out.println("ERROR. " + CellName[c] + " has zero perimeter! "); Sleep(); }
    double radius=halfPerimeter/3.1416;
    for (int t=1; t<=TerminalNumber[c]; t++) {
      //compute the center-terminal vector
      double tV=(double)(EuclideanDistance(TerminalOffset[c][t][0],TerminalOffset[c][t][1],cellCenterX,cellCenterY));
      //compute the terminal-tangential vector (orthogonal to center-terminal vector, and points counter-clockwise.
      //Normalized!
      double tTANX=-(TerminalOffset[c][t][1]-cellCenterY)/tV, tTANY=(TerminalOffset[c][t][0]-cellCenterX)/tV;
      //compute the projection of attract force on the terminal-tangential vector
      //this is a scalar. pointing counter-clockwise if positive, clockwise if negative.
      double fattTAN=TerminalForce[c][t][0]*tTANX+TerminalForce[c][t][1]*tTANY;  //already weighted between Chain and Silk
      //compensate the difference of center-terinal vector length, due to rectangle-to-circle mapping.
      fattTAN*=radius/tV;
      Terminal1DForce[c][t]=(int)(fattTAN/2*WeightCellAttractForce+Terminal1DFillingForce[c][t]*WeightCellFillingForce); 
      //System.out.println("$  Cell " + CellName[c] + " terminal " + TerminalName[c][t] + " tTANX/Y=" + tTANX + "/" + tTANY
      //                    + " fattTAN=" + fattTAN + " filling=" + Terminal1DFillingForce[c][t]);
    }
  }  

  void CalculateSoftBoundaryTerminalPosition(int c) {
    //
    CheckSoftBoundaryTerminal(c);
    DeterminePerimeterBin(c);
    CalculatePerimeterBinDensity(c);
    CalculatePerimeterBinFillingForce(c);
    CalculateSoftBoundaryTerminal1DForce(c);    
    //
    int perimeter=(CellSize[c][0]+CellSize[c][1])*2;
    int limit=java.lang.Math.min(SteinerMoveLimit, perimeter/4);
    int oneEdge=CellSize[c][0], twoEdge=CellSize[c][0]+CellSize[c][1], threeEdge=CellSize[c][0]*2+CellSize[c][1];
    for (int t=1; t<=TerminalNumber[c]; t++) {
      int offset=HowManyOffsetOfATerminalOnACellPerimeter(c,t);
      //slide terminal within limit
      if (Terminal1DForce[c][t]>limit) offset+=limit;
      else if (Terminal1DForce[c][t]<-limit) offset-=limit;
      else offset+=Terminal1DForce[c][t];
      //re-calculate terminalOffset
      offset=(offset+perimeter)%perimeter;
      if (offset<oneEdge)        { TerminalOffset[c][t][0]=offset; TerminalOffset[c][t][1]=0; }
      else if (offset<twoEdge)   { TerminalOffset[c][t][0]=oneEdge; TerminalOffset[c][t][1]=offset-oneEdge; }
      else if (offset<threeEdge) { TerminalOffset[c][t][0]=threeEdge-offset; TerminalOffset[c][t][1]=CellSize[c][1]; }
      else                       { TerminalOffset[c][t][0]=0; TerminalOffset[c][t][1]=offset-threeEdge; }
    }
  }

  void CalculateSoftBoundaryTerminalPosition() {
    for (int c=1; c<=CellNumber; c++) {
      if (CellType[c]=='d') CalculateSoftBoundaryTerminalPosition(c);
    }	
  }


 ///////////////////////////////////////////////////////////////////
 //ORIENTATION
 //
 ///////////////////////////////////////////////////////////////////
  void CalculateHardCellNewOrientationGain() {
    System.out.println("$  CALCULATE HARD CELL NEW ORIENTATION GAIN");
    for (int c=1; c<=CellNumber; c++) {
      if (CellType[c]!='C') continue;   //only handle hard cells.
      double dehFLIPX=0, dehFLIPY=0, dehROTPOS=0, dehROTNEG=0, dehROT180=0, dehTURNPOS=0, dehTURNNEG=0; //change in HOOK energy.
      double def=0;  //change in density.
      int cPX=CellPosition[c][0], cPY=CellPosition[c][1];
      int W=CellSize[c][0], H=CellSize[c][1];
      int ccPX=cPX+(W-H)/2, ccPY=cPY+(H-W)/2;   //new cell position after rotating 90 or 270 degree.
      for (int t=1; t<=TerminalNumber[c]; t++) {
        int tOX=TerminalOffset[c][t][0], tOY=TerminalOffset[c][t][1];
        int s=TerminalSilk[c][t];
        int tSL=TerminalSilkLength[c][t];
        double tSLR=((double)tSL)*((double)tSL);   //square of silkLength.
        int sPX=SteinerPosition[s][0], sPY=SteinerPosition[s][1];
        int tPX=TerminalPosition[c][t][0], tPY=TerminalPosition[c][t][1];
        //increase in HOOK energy. (NOTE: without division of 2)
        dehFLIPX+=EuclideanDistanceSquare((cPX+W-tOX),(cPY+tPY),sPX,sPY)-tSLR;
        dehFLIPY+=EuclideanDistanceSquare((cPX+tOX),(cPY+H-tOY),sPX,sPY)-tSLR;
        dehROT180+=EuclideanDistanceSquare((cPX+W-tOX),(cPY+H-tOY),sPX,sPY)-tSLR;
        dehROTPOS+=EuclideanDistanceSquare((ccPX+H-tOY),(ccPY+tOX),sPX,sPY)-tSLR;
        dehROTNEG+=EuclideanDistanceSquare((ccPX+tOY),(ccPY+W-tOX),sPX,sPY)-tSLR;
        dehTURNPOS+=EuclideanDistanceSquare((ccPX+H-tOY),(ccPY+W-tOX),sPX,sPY)-tSLR;
        dehTURNNEG+=EuclideanDistanceSquare((ccPX+tOY),(ccPY+tOX),sPX,sPY)-tSLR;
      }
      //change in density Energy.
      def=0.0;
      int b0left=CellBinBound[c][0], b0bottom=CellBinBound[c][1], b0right=CellBinBound[c][2], b0top=CellBinBound[c][3];
      int brleft=(ccPX-CellPosition[0][0])/BinSize+1, brbottom=(ccPY-CellPosition[0][1])/BinSize+1;
      int brright=(ccPX+H-CellPosition[0][0])/BinSize+1, brtop=(ccPY+W-CellPosition[0][1])/BinSize+1;
      if (brleft<1) brleft=1;
      if (brright>BinXNumber) brright=BinXNumber;
      if (brbottom<1) brbottom=1;
      if (brtop>BinYNumber) brtop=BinYNumber;
      //System.out.println("$  old BinBound=" + b0left + "," + b0bottom + "," + b0right + "," + b0top
      //                   + "  new BinBound=" + brleft + "," + brbottom + "," + brright + "," + brtop);
      if (brleft>b0left) {  //fat cell rotate to become slim cell
      	for (int bx=b0left; bx<brleft; bx++) for (int by=b0bottom; by<=b0top; by++) 
      	  def+=java.lang.Math.abs(BinCellDensity[bx][by]-2.0)-java.lang.Math.abs(BinCellDensity[bx][by]-1.0);
      	for (int bx=brright+1; bx<=b0right; bx++) for (int by=b0bottom; by<=b0top; by++) 
      	  def+=java.lang.Math.abs(BinCellDensity[bx][by]-2.0)-java.lang.Math.abs(BinCellDensity[bx][by]-1.0);
      	for (int bx=brleft; bx<=brright; bx++) for (int by=brbottom; by<b0bottom; by++)
      	  def+=java.lang.Math.abs(BinCellDensity[bx][by])-java.lang.Math.abs(BinCellDensity[bx][by]-1.0);
      	for (int bx=brleft; bx<=brright; bx++) for (int by=b0top+1; by<=brtop; by++)
      	  def+=java.lang.Math.abs(BinCellDensity[bx][by])-java.lang.Math.abs(BinCellDensity[bx][by]-1.0);
      }
      else if (brleft<b0left) { //slim cell rotate to become fat cell
      	for (int bx=brleft; bx<b0left; bx++) for (int by=brbottom; by<=brtop; by++) 
      	  def+=java.lang.Math.abs(BinCellDensity[bx][by])-java.lang.Math.abs(BinCellDensity[bx][by]-1.0);
      	for (int bx=b0right+1; bx<=brright; bx++) for (int by=brbottom; by<=brtop; by++) 
      	  def+=java.lang.Math.abs(BinCellDensity[bx][by])-java.lang.Math.abs(BinCellDensity[bx][by]-1.0);
      	for (int bx=b0left; bx<=b0right; bx++) for (int by=b0bottom; by<brbottom; by++)
      	  def+=java.lang.Math.abs(BinCellDensity[bx][by]-2.0)-java.lang.Math.abs(BinCellDensity[bx][by]-1.0);
      	for (int bx=b0left; bx<=b0right; bx++) for (int by=brtop+1; by<=b0top; by++)
      	  def+=java.lang.Math.abs(BinCellDensity[bx][by]-2.0)-java.lang.Math.abs(BinCellDensity[bx][by]-1.0);
      }
      def*=(BinSize*BinSize);
      //change in TOTAL Energy, gain is the negative of energy increasing.
      double wghtH=-WeightCellAttractForce;   //no need to /2, 'cos def also has a /2 term
      double wghtF=-WeightCellFillingForce;
      CellFlipXGain[c]    =-wghtH*dehFLIPX;
      CellFlipYGain[c]    =-wghtH*dehFLIPY;
      CellRotatePosGain[c]=-wghtH*dehROTPOS-wghtF*def;
      CellRotateNegGain[c]=-wghtH*dehROTNEG-wghtF*def;
      CellRotate180Gain[c]=-wghtH*dehROT180;
      CellTurnPosGain[c]  =-wghtH*dehTURNPOS-wghtF*def; 
      CellTurnNegGain[c]  =-wghtH*dehTURNNEG-wghtF*def;
    }
  }
  
  void CalculateHardCellOrientation() {  //return true when one cell changes orientation, false if none.
    System.out.println("$ CALCULATE HARD CELL ORIENTATION");
    CalculateHardCellNewOrientationGain();
   //determine which cell to flip or rotate.
   //RULE 1: only one cell is allowed to flip/rotate
    double BestGain=0; int BestCell=0; char BestOrientation='n';
    for (int c=1; c<=CellNumber; c++) {
      if (CellType[c]!='C') continue;    //skip non-hard cell.
      if (CellFlipXGain[c]>BestGain) { BestGain=CellFlipXGain[c]; BestCell=c; BestOrientation='h'; }
      if (CellFlipYGain[c]>BestGain) { BestGain=CellFlipYGain[c]; BestCell=c; BestOrientation='v'; }
      if (CellRotateNegGain[c]>BestGain) { BestGain=CellRotateNegGain[c]; BestCell=c; BestOrientation='c'; }
      if (CellRotatePosGain[c]>BestGain) { BestGain=CellRotatePosGain[c]; BestCell=c; BestOrientation='w'; }
      if (CellRotate180Gain[c]>BestGain) { BestGain=CellRotate180Gain[c]; BestCell=c; BestOrientation='o'; }
      if (CellTurnPosGain[c]>BestGain) { BestGain=CellTurnPosGain[c]; BestCell=c; BestOrientation='t'; }
      if (CellTurnNegGain[c]>BestGain) { BestGain=CellTurnNegGain[c]; BestCell=c; BestOrientation='r'; }
    }
    if (BestGain>0) {       
      System.out.println("$  Cell " + CellName[BestCell] + " with gain of " + BestGain + " with action of " + BestOrientation);
      String orientationTransferTable = new String(
                           //nhvcwotr
                            "nhvwestr"  //n
                          + "wrtsnehv"  //w
                          + "svhewnrt"  //s
                          + "etrnswvh"  //e
                          + "hnstrvwe"  //h
                          + "vsnrthew"  //v
                          + "tewvhrns"  //t
                          + "rwehvtsn"  //r
                         );                         
      int coIndex=0, oaIndex=0;
      switch (CellOrientation[BestCell]) {
        case 'n': coIndex=0; break;
        case 'w': coIndex=1; break;
        case 's': coIndex=2; break;
        case 'e': coIndex=3; break;
        case 'h': coIndex=4; break;
        case 'v': coIndex=5; break;
        case 't': coIndex=6; break;
        case 'r': coIndex=7; break;
        default: System.out.println("ERROR. Unrecognized CellOrientation " + CellName[BestCell]); Sleep();
      }
      switch (BestOrientation) {
        case 'n': oaIndex=0; break;
        case 'h': oaIndex=1; break;
        case 'v': oaIndex=2; break;
        case 'c': oaIndex=3; break;
        case 'w': oaIndex=4; break;
        case 'o': oaIndex=5; break;
        case 't': oaIndex=6; break;
        case 'r': oaIndex=7; break;
        default: System.out.println("ERROR. Unrecognized Best Orientation Action " + BestOrientation); Sleep();
      }
      CellOrientation[BestCell]=orientationTransferTable.charAt(coIndex*8+oaIndex);
      UpdateACell(BestCell);
    }
    else {
      System.out.println("$  No cell changes orientation.");
    }
  }











//////////////////////////////////////////////////////////////////////////////
//CLEAN UP
//USING MOSEK QUADRATIC SOLVER
//
//////////////////////////////////////////////////////////////////////////////
  void Compactor(char direction, int maxOver, boolean needWireBalancing) {
   //PARAMETER
    int MinSpacing=0;
    double compactionWireRate=0.05;
    double wireBalanceWeight=0.0;
    int dir=(direction=='x')?0:1;
    double starSquareMatrix[][]; starSquareMatrix=new double[CellLimit+1][CellLimit+1];
    double starLinearVector[]; starLinearVector=new double[CellLimit+1];
    //needed in this procedure.
    int SilkTerminalNumber[], SilkTerminalCell[][], SilkTerminalTerminal[][];  //dual-directional pointer
    SilkTerminalNumber=new int[SilkNumber+1]; 
    SilkTerminalCell=new int[SilkNumber+1][60]; SilkTerminalTerminal = new int[SilkNumber+1][60];
    for (int s=1; s<=SilkNumber; s++) SilkTerminalNumber[s]=0;
    for (int c=0; c<=CellNumber; c++) {
      for (int t=1; t<=TerminalNumber[c]; t++) {
      	int s=TerminalSilk[c][t];
      	SilkTerminalNumber[s]++; 
      	SilkTerminalCell[s][SilkTerminalNumber[s]]=c; SilkTerminalTerminal[s][SilkTerminalNumber[s]]=t;
      }
    }
    for (int s=1; s<=SilkNumber; s++) { 
      if (SilkTerminalNumber[s]==0) { System.out.println("ERROR. Silk " + SilkName[s] + " has zero connection."); Sleep(); } 
    }
   ////////////////////////////////////////////////////////////////////////////////////////
   // GENERATE MPS FILE FOR MOSEK TO WORK WITH 
   //
    try {
      OutputStream mps = new FileOutputStream(AgeFilePrefix + ".mps");
     //////////////////////////////////////////////////////////////
     //generate spacing constraint graph
     //  spacing[ca][cb].  0: no constraint; 
     //                   >0: constraint from ca to cb, thus generate Pcb-Pca>wa; 
     //                   -1: constraint exist from cb to ca, constraint can be found in spacing[cb][ca].
      int spacing[][]; spacing=new int[CellLimit+1][CellLimit+1];
      System.out.println("#(0)BUILD SPACING CONSTRAINT GRAPH. REDUNDANT CONSTRAINTS EXIST.");
      for (int c1=1; c1<CellNumber; c1++) {
      	int c1xc=CellPosition[c1][dir]+CellSize[c1][dir]/2;
      	int c1y0=CellPosition[c1][1-dir]; int c1y1=c1y0+CellSize[c1][1-dir];
      	for (int c2=c1+1; c2<=CellNumber; c2++) {
          int c2xc=CellPosition[c2][dir]+CellSize[c2][dir]/2;
      	  int c2y0=CellPosition[c2][1-dir]; int c2y1=c2y0+CellSize[c2][1-dir];
      	  if (c2y1<c1y0+maxOver | c2y0>c1y1-maxOver) continue;  //no y-overlapping
      	  if (c1xc<c2xc) { spacing[c1][c2]=CellSize[c1][dir]; spacing[c2][c1]=-1; }
      	  else           { spacing[c2][c1]=CellSize[c2][dir]; spacing[c1][c2]=-1; }
      	}
      }
     //////////////////////////////////////////////////////////////
     //generate star net model
      if (needWireBalancing) {
       //////////////////////////////////////////////////////////////
       //determine wireBalanceWeight
       //measure current squared wire length sum
       //compare it to the current chip width, make the ratio be 1e-4
        System.out.println("#(1)DETERMINE WIRE BALANCE WEIGHT.");
        double totalSquaredWireLength=0.0;
        for (int s=1; s<=SilkNumber; s++) {
      	  int starX=0;
      	  for (int sT=1; sT<=SilkTerminalNumber[s]; sT++) {
      	    int c=SilkTerminalCell[s][sT], t=SilkTerminalTerminal[s][sT];
      	    starX+=CellPosition[c][dir]+TerminalOffset[c][t][dir];
      	  }
      	  starX/=SilkTerminalNumber[s];
      	  for (int sT=1; sT<=SilkTerminalNumber[s]; sT++) {
      	    int c=SilkTerminalCell[s][sT], t=SilkTerminalTerminal[s][sT];
      	    int tx=CellPosition[c][dir]+TerminalOffset[c][t][dir];
      	    double wx=(double)(tx-starX);
      	    totalSquaredWireLength+=wx*wx;
      	  }
        }
        wireBalanceWeight=CellSize[0][dir]/totalSquaredWireLength*compactionWireRate;
        System.out.println("#  the original layout, in " + direction + " direction, has width " 
                            + CellSize[0][dir] + " and total squared wire length " + totalSquaredWireLength);
        System.out.println("#  compationWireRate is " + compactionWireRate + " wireBalanceWeight is " + wireBalanceWeight);
       //annotate the quadratic coefficient matrix and the linear coefficient matrix
       //index 0 store xR terms
       //see the course report for details.
        System.out.println("#(2)BUILD STAR NET MODEL AND CREATE THE COEFFICIENT MATRIX/VECTOR.");
        for (int s=1; s<=SilkNumber; s++) {
          //System.out.println("#  silk" + SilkName[s]);
      	  int terminalCell[]; terminalCell=new int[SilkTerminalNumber[s]+1];
      	  double terminalOffset[]; terminalOffset=new double[SilkTerminalNumber[s]+1];
      	  double beta=0; 
      	  int t=0, p=0;
      	  for (int stA=1; stA<=SilkTerminalNumber[s]; stA++) {
      	    int cA=SilkTerminalCell[s][stA], tA=SilkTerminalTerminal[s][stA];
      	    if (cA>0) {
      	      t++;
      	      terminalCell[t]=cA;
      	      terminalOffset[t]=(double)(TerminalOffset[cA][tA][dir]);
              //System.out.println("#    terminal " + CellName[cA] + "/" + TerminalName[cA][tA]);
      	    }
      	    else {
      	      beta=((double)(TerminalOffset[0][tA][dir]))/((double)(CellSize[0][dir]));
      	      p=1;	
              //System.out.println("#    pad      " + CellName[cA] + "/" + TerminalName[cA][tA]);
      	    }
      	  }
      	  //System.out.println("#    t=" + t + " p=" + p);
      	  //fill up the coefficient matrix
          double tmp1=1.0-1.0/((double)(t+p));
          for (int i=1; i<=t; i++) starSquareMatrix[terminalCell[i]][terminalCell[i]]+=tmp1;
      	  if (p!=0) starSquareMatrix[0][0]+=(t*beta*beta)/((double)(t+1));
      	  double tmp2=-1.0/((double)(t+p));
      	  for (int i=1; i<t; i++) {
      	    for (int j=i+1; j<=t; j++) {
      	      starSquareMatrix[terminalCell[i]][terminalCell[j]]+=tmp2;
      	      starSquareMatrix[terminalCell[j]][terminalCell[i]]+=tmp2;
      	    }
      	  }
      	  if (p!=0) {
            double tmp3=-((double)p)*beta/((double)(t+p));
      	    for (int i=1; i<=t; i++) { 
      	      starSquareMatrix[0][terminalCell[i]]+=tmp3;
      	      starSquareMatrix[terminalCell[i]][0]+=tmp3;
      	    }
      	  }
          //fill up the coefficient vector
      	  double xts=0;
      	  for (int i=1; i<=t; i++) xts+=terminalOffset[i];
      	  xts/=t;
      	  double tmp5=((double)t)/((double)(t+p))*xts;
      	  for (int i=1; i<=t; i++) starLinearVector[terminalCell[i]]+=2*(terminalOffset[i]-tmp5);
      	  if (p!=0) starLinearVector[0]+=2*tmp5*p*beta;
        }
      }
     //////////////////////////////////////////////////////////////
     //write mps file
      System.out.println("#(3)WRITE MPS FILE.");
      WriteALineToFile(("NAME " + CellName[0] + "\r\n"), mps);
      //row definition
      WriteALineToFile(("ROWS\r\n"), mps);
      WriteALineToFile((" N  obj\r\n"), mps);
      for (int c2=1; c2<=CellNumber; c2++) WriteALineToFile((" G  sp_L_" + c2 + "\r\n"), mps);
      for (int c1=1; c1<=CellNumber; c1++) {
      	for (int c2=1; c2<=CellNumber; c2++) {
      	  if (spacing[c1][c2]>0) WriteALineToFile((" G  sp_" + c1 + "_" + c2 + "\r\n"), mps);
      	}
      }
      for (int c1=1; c1<=CellNumber; c1++) WriteALineToFile((" G  sp_" + c1 + "_R\r\n"), mps);
      //column definition
      WriteALineToFile(("COLUMNS\r\n"), mps);
      for (int c1=1; c1<=CellNumber; c1++) {
        String st1=new String("    x" + c1 + "                 ");
        String st2=new String(st1.substring(0,14));
        if (needWireBalancing & starLinearVector[c1]!=0) WriteALineToFile((st2 + "obj       " + (starLinearVector[c1]*wireBalanceWeight) + "\r\n"), mps);
        String st3=new String(st2.concat("sp_L_" + c1 + "                ").substring(0,24).concat("1"));
        WriteALineToFile((st3 + "\r\n"), mps);
      	for (int c2=1; c2<=CellNumber; c2++) {
      	  if (spacing[c1][c2]!=0) {
      	    String st4=st2.concat("sp_" + ((spacing[c1][c2]>0)?(c1 + "_" + c2):(c2 + "_" + c1)) + "                   ");
      	    String st5=new String(st4.substring(0,24));
      	    st5=st5.concat(""+((spacing[c1][c2]>0)?("-1"):("1")));
      	    WriteALineToFile((st5 + "\r\n"), mps);
      	  }
      	}
        String st7=new String(st2.concat("sp_" + c1 + "_R                 ").substring(0,24).concat("-1"));
        WriteALineToFile((st7 + "\r\n"), mps);
      }
      WriteALineToFile(("    xR        obj       " + (1.0+(needWireBalancing?(starLinearVector[0]*wireBalanceWeight):0)) + "\r\n"), mps);
      for (int c1=1; c1<=CellNumber; c1++) {
        String st1=new String("    xR        sp_" + c1 + "_R                ");
        String st2=new String(st1.substring(0,24).concat("1"));
        WriteALineToFile((st2 + "\r\n"), mps);
      }
      //rhs
      WriteALineToFile(("RHS\r\n"), mps);
      for (int c2=1; c2<=CellNumber; c2++) {
        String st1=new String("    rhs       sp_L_" + c2 + "                ");
        String st2=new String(st1.substring(0,24).concat("" + MinSpacing));
        WriteALineToFile((st2 + "\r\n"), mps);
      }
      for (int c1=1; c1<=CellNumber; c1++) {
      	for (int c2=1; c2<=CellNumber; c2++) {
      	  if (spacing[c1][c2]!=0) {
      	    String st1=new String("    rhs                         ");
      	    String st2=new String(st1.substring(0,14));
      	    st2=st2.concat("sp_" + ((spacing[c1][c2]>0)?(c1 + "_" + c2):(c2 + "_" + c1)) + "                   ");
      	    String st3=new String(st2.substring(0,24));
      	    st3=st3.concat(""+((spacing[c1][c2]>0)?(spacing[c1][c2]+MinSpacing):(spacing[c2][c1]+MinSpacing)));
      	    WriteALineToFile((st3 + "\r\n"), mps);
      	  }
      	}
      }
      for (int c1=1; c1<=CellNumber; c1++) {
        String st1=new String("    rhs       sp_" + c1 + "_R                ");
        String st2=new String(st1.substring(0,24).concat(""+(CellSize[c1][dir]+MinSpacing)));
        WriteALineToFile((st2 + "\r\n"), mps);
      }
      //qsection
      if (needWireBalancing) {
        WriteALineToFile(("QSECTION      obj\r\n"), mps);
        if (starSquareMatrix[0][0]!=0) {
          WriteALineToFile(("    xR        xR        " + (starSquareMatrix[0][0]*wireBalanceWeight) + "\r\n"), mps);
        }
        for (int c2=1; c2<=CellNumber; c2++) {
      	  if (starSquareMatrix[0][c2]!=0) {
            String st1=new String("    xR              ");
            String st2=new String(st1.substring(0,14).concat("x" + c2 + "                 "));
            String st3=new String(st2.substring(0,24).concat("" + (starSquareMatrix[0][c2]*wireBalanceWeight)));
            WriteALineToFile((st3 + "\r\n"), mps);
          }
        }
        for (int c1=1; c1<=CellNumber; c1++) {
          String st1=new String("    x" + c1 + "                ");
      	  for (int c2=c1; c2<=CellNumber; c2++) {
      	    if (starSquareMatrix[c1][c2]!=0) {
              String st2=new String(st1.substring(0,14).concat("x" + c2 + "                 "));
              String st3=new String(st2.substring(0,24).concat("" + (starSquareMatrix[c1][c2]*2*wireBalanceWeight)));
              WriteALineToFile((st3 + "\r\n"), mps);
            }
      	  }
        }
      }
      //end
      WriteALineToFile(("ENDATA\r\n"), mps);
     //////////////////////////////////////////////////////////////
     //over
      mps.close();
      System.out.println("#(4)FINISH MPS FILE PREPARATION!");
    }
    catch (FileNotFoundException e) { System.out.println("ERROR. File not Found!"); }
    catch (IOException e) { System.out.println("ERROR. IO Exception Occurs!"); }
    finally { }
   ////////////////////////////////////////////////////////////////////////////////////////
   // RUN MOSEK
   //
    Runtime r=Runtime.getRuntime();
    Process p=null;
    String cmd[]={"mosek", (AgeFilePrefix + ".mps")};
    try {
      p=r.exec(cmd);
      //p.waitFor();
      System.out.println("#(10)MOSEK DONE.");
    }
    catch(Exception e) { System.out.println("#ERROR executing " + cmd[0]); }
   ////////////////////////////////////////////////////////////////////////////////////////
   // READ IN SOLUTION FILE FROM MOSEK AND WRITE TO AGEOUT ORG
   //
   // Read in mosek results
    for (int i=1; i<=1000000000; i++) {};
    System.out.println("#(21)READ MOSEK RESULTS.");
    try {
      int LineNumber=0;
      char InputChar;
      String InputLine = new String();
      InputStream f=new FileInputStream(AgeFilePrefix + ".sol");
      String[] LineField;
      LineField = new String[20];
      boolean effective=false;
      while (f.available()!=0) {
        InputChar=(char)f.read();
        if (InputChar!='\n') InputLine+=InputChar;
        else {
          LineNumber ++; // System.out.println(" read line " + LineNumber + " \n");
          StringTokenizer st = new StringTokenizer(InputLine, " \r\n\t");
          int Token = 0;
          while (st.hasMoreTokens()) { Token++; LineField[Token] = st.nextToken(); }
          if (Token == 0) { } // do nothing.
          else {
            if (effective==false) {
              if (LineField[1].equals("VARIABLES")) effective=true;
            }
            else {
              if (LineField[2].charAt(0)=='x') {
              	switch (LineField[2].charAt(1)) {
              	  case 'R': CellPosition[0][dir]=0;
              	            int oldChipSize=CellSize[0][dir];
              	            CellSize[0][dir]=(int)(StringToDouble(LineField[4]))-CellPosition[0][dir];
              	            double ratio=((double)(CellSize[0][dir]))/((double)oldChipSize);
              	            for (int t=1; t<=TerminalNumber[0]; t++) {
              	              TerminalOffset[0][t][dir]=(int)(TerminalOffset[0][t][dir]*ratio);
              	            }
              	            break;
              	  default : int c=StringToInt(LineField[2].substring(1));
              	            //System.out.println("# reading cell no. " + c);
              	            CellPosition[c][dir]=(int)(StringToDouble(LineField[4]));
              	            break;
              	}
              }
            }
          }
          InputLine = "";
        }
      }
      f.close();
    }
    catch (FileNotFoundException e) { System.out.println("File not Found!"); Sleep(); }
    catch (IOException e) { System.out.println("IO Exception Occurs!"); Sleep(); }
    finally { }
   ////////////////////////////////////////////////////////////////////////////////////////
   // MOVE STEINER TO THE CENTER OF GRAVITY OF ITS TERMINALS
    for (int s=1; s<=SilkNumber; s++) {
      int tn=SilkTerminalNumber[s];	
      int sx=0, sy=0;
      for (int tt=1; tt<=tn; tt++) {
      	int c=SilkTerminalCell[s][tt], t=SilkTerminalTerminal[s][tt];
      	sx+=CellPosition[c][0]+TerminalOffset[c][t][0]; sy+=CellPosition[c][1]+TerminalOffset[c][t][1]; 
      }	
      SteinerPosition[s][0]=sx/tn; SteinerPosition[s][1]=sy/tn; 
    }
  }

  void Compaction2D() {
    boolean needWireBalancing=true;
    int maxOver=10;
    System.out.println("\r\n&COMPACTION" + ((needWireBalancing)?(" WITH"):(" WITHOUT")) + " WIRE BALANCING");
    //x compaction
    System.out.println("&XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX");
    //age++;
    //Age(age);
    Compactor('x', maxOver, needWireBalancing);
    //Write();
    //y compaction
    System.out.println("&YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY");
    //age++;
    //Age(age);
    Compactor('y', maxOver, needWireBalancing);
    //Write();
  }































//END OF CLASS BioMechanics
}


  


